/* ScummVM - Graphic Adventure Engine
 *
 * ScummVM is the legal property of its developers, whose names
 * are too numerous to list here. Please refer to the COPYRIGHT
 * file distributed with this source distribution.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.

 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.

 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 *
 */

#include "audio/decoders/cdda.h"

#include "common/debug.h"
#include "common/ptr.h"
#include "common/stream.h"
#include "common/textconsole.h"
#include "common/util.h"

#include "audio/audiostream.h"

namespace Audio {


#pragma mark -
#pragma mark --- CDDA stream ---
#pragma mark -

class CDDAStream : public SeekableAudioStream {
protected:
	Common::SeekableReadStream *_stream;
	bool _disposeAfterUse;
	uint8 _shiftLeft;
	uint8 _shiftRight;
	uint32 _pos;

public:
	CDDAStream(Common::SeekableReadStream *stream, bool dispose);
	virtual ~CDDAStream();
	
	int readBuffer(int16 *buffer, const int numSamples);
	bool isStereo() const { return 1; }
	int getRate() const { return 44100; }
	bool endOfData() const { return _stream->eos(); }
	bool seek(const Timestamp &where);
	Timestamp getLength() const { return 0; }
};

CDDAStream::CDDAStream(Common::SeekableReadStream *stream, bool dispose) :
	_stream(stream),
	_disposeAfterUse(dispose) {

	_shiftLeft = 0;
	_shiftRight = 0;
	_pos = 800;
	_stream->seek(800, SEEK_SET);
}

CDDAStream::~CDDAStream() {
	if (_disposeAfterUse)
		delete _stream;
}

bool CDDAStream::seek(const Timestamp &where) {

	const uint32 seekSample = convertTimeToStreamPos(where, getRate(), isStereo()).totalNumberOfFrames();
	uint32 blocks = seekSample / 1176;

	_pos = 800 + blocks + seekSample;
	return _stream->seek(_pos, SEEK_SET);
}

int CDDAStream::readBuffer(int16 *buffer, const int numSamples) {
	int samples;

	for (samples = 0 ; samples < numSamples && !_stream->eos() ; ) {
		if ( !((_pos - 800) % 1177) ) {
			_shiftRight = _stream->readByte();
			_shiftLeft = _shiftRight >> 4;
			_shiftRight = _shiftRight & 0x0F;
			_pos++;
		}
		buffer[samples++] = _stream->readSByte() << _shiftLeft;
		buffer[samples++] = _stream->readSByte() << _shiftRight;
		_pos += 2;
	}
	return samples;
}

#pragma mark -
#pragma mark --- CDDA factory functions ---
#pragma mark -

SeekableAudioStream *makeCDDAStream(
	Common::SeekableReadStream *stream,
	DisposeAfterUse::Flag disposeAfterUse) {
	SeekableAudioStream *s = new CDDAStream(stream, disposeAfterUse);
	if (s && s->endOfData()) {
		delete s;
		return 0;
	} else {
		return s;
	}
	return 0;
}

} // End of namespace Audio
