package org.iphsoft.simon1;

import java.util.ArrayList;
import java.util.List;
import java.util.Locale;
import java.util.TreeSet;

import org.iphsoft.simon1.stats.StatisticsManager;
import org.iphsoft.simon1.util.CsvUtils;
import org.iphsoft.simon1.util.DeviceIdentifier;
import org.iphsoft.simon1.util.DeviceIdentifier.DeviceModel;
import org.iphsoft.simon1.util.MyLog;

import android.content.Context;
import android.content.SharedPreferences;
import android.content.SharedPreferences.Editor;
import android.content.pm.PackageManager;
import android.os.Build;
import android.preference.ListPreference;
import android.preference.PreferenceManager;

public class AndroidPortAdditions {

	public enum GraphicModeBehavior
	{
		SOFTWARE_SCALER, ULTRA_HARDWARE_SCALER, HARDWARE_SCALER, TEST_HARDWARE_SCALER, LQ_HARDWARE_SCALER, TEST_LQ_HARDWARE_SCALER, ORIGINAL
	}

	public enum AppStore
	{
		GOOGLE_PLAY, AMAZON, SAMSUNG
	}

	public enum GameType
	{
		SIMON1(0), SIMON2(1), FOTAQ(2), INDY_FOA(3), T7G(4), IHNM(5);

		public int getValue()
		{
			return mValue;
		}

		public static GameType fromValue(int value)
		{
			for (GameType type : GameType.values())
			{
				if (type.getValue() == value)
				{
					return type;
				}
			}

			return null;
		}

		private GameType(int value)
		{
			mValue = value;
		}

		private int mValue;
	}

	public static GameType GAME_TYPE = GameType.fromValue(ScummVM.getCurrentGameType());

	public static final boolean RELEASE_VERSION = true;

	// Used for testing shaders - allows to set external shader files
	public static boolean USE_SHADER_FILES;

	// Testing features for achievements
	public static boolean TEST_ACHIEVEMENTS;

	// Whether to always do a shader benchmark
	public static boolean TEST_SHADER_BENCHMARK;

	// Whether to allow modifying game files for testing
	public static boolean ALLOW_GAME_FILE_MODIFICATION;

	// Whether to show debug toasts
	public static boolean SHOW_DEBUG_TOASTS;

	public static boolean T7G_ALWAYS_OPEN_HOUSE;

	public static boolean MENU_FORCE_PAUSE;

	public static boolean EXTRAS_EXTERNAL_JSON_FILE;
	
	public static boolean HIDE_GAME_START_LOGO;


	static
	{
		if (RELEASE_VERSION)
		{
			USE_SHADER_FILES = false;
			TEST_ACHIEVEMENTS = false;
			TEST_SHADER_BENCHMARK = false;
			ALLOW_GAME_FILE_MODIFICATION = false;
			SHOW_DEBUG_TOASTS = false;
			T7G_ALWAYS_OPEN_HOUSE = false;
			MENU_FORCE_PAUSE = false;
			EXTRAS_EXTERNAL_JSON_FILE = false;
			HIDE_GAME_START_LOGO = false;

		} else
		{
			switch (AndroidPortAdditions.GAME_TYPE)
			{
			case SIMON1:
			case SIMON2:
			case IHNM:
				USE_SHADER_FILES = true;
				TEST_ACHIEVEMENTS = false;
				TEST_SHADER_BENCHMARK = false;
				ALLOW_GAME_FILE_MODIFICATION = false;
				SHOW_DEBUG_TOASTS = false;
				T7G_ALWAYS_OPEN_HOUSE = false;
				MENU_FORCE_PAUSE = false;
				EXTRAS_EXTERNAL_JSON_FILE = false;
				HIDE_GAME_START_LOGO = false;
				break;
			case FOTAQ:
				USE_SHADER_FILES = true;
				TEST_ACHIEVEMENTS = false;
				TEST_SHADER_BENCHMARK = false;
				ALLOW_GAME_FILE_MODIFICATION = true;
				SHOW_DEBUG_TOASTS = false;
				T7G_ALWAYS_OPEN_HOUSE = false;
				MENU_FORCE_PAUSE = false;
				EXTRAS_EXTERNAL_JSON_FILE = true;
				HIDE_GAME_START_LOGO = false;
				break;
			case T7G:
				USE_SHADER_FILES = true;
				TEST_ACHIEVEMENTS = false;
				TEST_SHADER_BENCHMARK = false;
				ALLOW_GAME_FILE_MODIFICATION = true;
				SHOW_DEBUG_TOASTS = true;
				T7G_ALWAYS_OPEN_HOUSE = true;
				MENU_FORCE_PAUSE = false;
				EXTRAS_EXTERNAL_JSON_FILE = true;
				HIDE_GAME_START_LOGO = false;
				break;
			}

		}
	}

	public static final AppStore APP_STORE = AppStore.GOOGLE_PLAY;

	// True if the game files are inside the APK
	public static final boolean STANDALONE_APK = (APP_STORE == AppStore.GOOGLE_PLAY) ? false : true;

	public static final boolean SUPPORTS_AUTO_SAVE = (GAME_TYPE == GameType.T7G
			|| GAME_TYPE == GameType.IHNM || GAME_TYPE == GameType.FOTAQ);

	public static final boolean SUPPORTS_SOFTWARE_SCALER = (GAME_TYPE != GameType.T7G && GAME_TYPE != GameType.IHNM);

	public static final boolean SUPPORTS_FACEBOOK_ACHIEVEMENTS = (GAME_TYPE == GameType.T7G);
	
	public static final boolean HAS_TUTORIAL = (GAME_TYPE == GameType.SIMON1 || GAME_TYPE == GameType.SIMON2 || GAME_TYPE == GameType.FOTAQ);

	public static final boolean HAS_EXTRAS = (GAME_TYPE == GameType.T7G || GAME_TYPE == GameType.FOTAQ || GAME_TYPE == GameType.IHNM);

	public static final boolean HAS_BACKGROUND_MUSIC = (GAME_TYPE == GameType.T7G || GAME_TYPE == GameType.IHNM || GAME_TYPE == GameType.FOTAQ || GAME_TYPE == GameType.SIMON1 || GAME_TYPE == GameType.SIMON2);


	private static String sTempGameFileDir;
	static
	{
		switch (GAME_TYPE)
		{
		case SIMON1:
			sTempGameFileDir = "/simon1";
			break;
		case SIMON2:
			sTempGameFileDir = "/simon2";
			break;
		case FOTAQ:
			sTempGameFileDir = "/fotaq";
			break;
		case INDY_FOA:
			sTempGameFileDir = "/indy";
			break;
		case T7G:
			sTempGameFileDir = "/t7g";
			break;
		case IHNM:
			sTempGameFileDir = "/ihnmaims";
			break;
		}
	}

	public static final String GAME_FILE_DIR = sTempGameFileDir;

	public enum MusicOption
	{
		NONE("none"), ORIGINAL_RECORDING("original_recording"), ADLIB("original"), EAS(
				"original_eas"), MT32("original_mt32"), ENHANCED("enhanced");

		public String getValue()
		{
			return mValue;
		}

		public static MusicOption fromValue(String value)
		{
			for (MusicOption type : MusicOption.values())
			{
				if (type.getValue().equals(value))
				{
					return type;
				}
			}

			return null;
		}

		private MusicOption(String value)
		{
			mValue = value;
		}

		private String mValue;
	}

	public static final int GAME_OPTION_NONE = 0;
	public static final int GAME_OPTION_SAVE = 1;
	public static final int GAME_OPTION_LOAD = 2;
	public static final int GAME_OPTION_EXIT = 3;

	public static final int SCALING_OPTION_SHADER = 0;
	public static final int SCALING_OPTION_SOFT = 1;
	public static final int SCALING_OPTION_NONE = 2;
	public static final int SCALING_OPTION_LQ_SHADER = 3;
	public static final int SCALING_OPTION_ULTRA_SHADER = 4;

	public static final int GAME_EVENT_SCALER_FALLBACK = 0;
	public static final int GAME_EVENT_SHOULD_TEST_SHADER = 1;
	public static final int GAME_EVENT_SHADER_TEST_SUCCESS = 2;
	public static final int GAME_EVENT_SHADER_TEST_FAILURE = 3;
	public static final int GAME_EVENT_LOAD_SUCCESS = 4;
	public static final int GAME_EVENT_LOAD_FAILURE = 5;
	public static final int GAME_EVENT_SAVE_SUCCESS = 6;
	public static final int GAME_EVENT_SAVE_FAILURE = 7;
	public static final int GAME_EVENT_SAVE_CANCELLED = 8;
	public static final int GAME_EVENT_HEBREW_TITLE_GONE = 9;
	public static final int GAME_EVENT_POSTCARD_SHOWN = 10;
	public static final int GAME_EVENT_SHOW_MENU = 11;
	public static final int GAME_EVENT_SCALER_LQ_FALLBACK = 12;
	public static final int GAME_EVENT_USE_ULTRA_MODE = 13;
	public static final int GAME_EVENT_HIDE_NEW_ACTION_BAR = 14;
	public static final int GAME_EVENT_SHOW_NEW_ACTION_BAR = 15;
	public static final int GAME_EVENT_SKIP_INTRO = 16;
	public static final int GAME_EVENT_ENABLE_OPEN_HOUSE = 17;
	public static final int GAME_EVENT_START_GAME_IN_OPEN_HOUSE = 18;
	public static final int GAME_EVENT_T7G_CANS_PUZZLE_REACHED = 19;
	public static final int GAME_EVENT_SKIP_CAN_PUZZLE = 20;
	public static final int GAME_EVENT_T7G_BEDSPREAD_PUZZLE_REACHED = 21;
	public static final int GAME_EVENT_SKIP_BEDSPREAD_PUZZLE = 22;
	public static final int GAME_EVENT_T7G_MICROSCOPE_PUZZLE = 23;
	public static final int GAME_EVENT_T7G_HEART_PUZZLE_REACHED = 24;
	public static final int GAME_EVENT_SKIP_HEART_PUZZLE = 25;
	public static final int GAME_EVENT_T7G_INTRO_LOGO_END = 26;
	public static final int GAME_EVENT_FOTAQ_INTRO_LOGO_END = 27;
	public static final int GAME_EVENT_IHNM_INTRO_LOGO_END = 28;
	public static final int GAME_EVENT_T7G_BISHOPS_PUZZLE = 29;
	public static final int GAME_EVENT_SET_FORCE_AUTOSAVING = 30;



	public static synchronized AndroidPortAdditions instance()
	{
		if (sInstance == null)
			sInstance = new AndroidPortAdditions();

		return sInstance;
	}

	public String getAppName()
	{
		Context appContext = ScummVMApplication.getContext();
		PackageManager pm = appContext.getPackageManager();
		CharSequence appName = pm.getApplicationLabel(appContext.getApplicationInfo());
		return appName.toString();
	}

	public void setPreferenceDefaults()
	{
		Context context = ScummVMApplication.getContext();

		SharedPreferences prefs = PreferenceManager.getDefaultSharedPreferences(context);

		if (AndroidPortAdditions.instance().getFirstUse())
		{
			// If this is the first use, set the default language according to
			// device locale
			String defaultLanguagePref = "en";

			// Languages appearing in all games
			if (checkLocale("de"))
				defaultLanguagePref = "de";
			if (checkLocale("fr"))
				defaultLanguagePref = "fr";

			if (AndroidPortAdditions.GAME_TYPE == GameType.SIMON1)
			{
				// Languages appearing in Simon 1
				if (checkLocale("it"))
					defaultLanguagePref = "it";
				if (checkLocale("es"))
					defaultLanguagePref = "es";
				if (checkLocale("he") || checkLocale("iw"))
					defaultLanguagePref = "he";
			} else if (AndroidPortAdditions.GAME_TYPE == GameType.SIMON2)
			{
				// Languages appearing in Simon 2
				if (checkLocale("it"))
					defaultLanguagePref = "it";
				if (checkLocale("cs"))
					defaultLanguagePref = "cz";
				if (checkLocale("pl"))
					defaultLanguagePref = "pl";
				if (checkLocale("es"))
					defaultLanguagePref = "es";
				if (checkLocale("he") || checkLocale("iw"))
					defaultLanguagePref = "he";
			} else if (AndroidPortAdditions.GAME_TYPE == GameType.T7G)
			{
				// Languages appearing in T7G
				if (checkLocale("ru"))
					defaultLanguagePref = "ru";
			} else if (AndroidPortAdditions.GAME_TYPE == GameType.FOTAQ)
			{
				// Languages appearing in FOTAQ
				if (checkLocale("it"))
					defaultLanguagePref = "it";
				if (checkLocale("es"))
					defaultLanguagePref = "es";
				if (checkLocale("he") || checkLocale("iw"))
					defaultLanguagePref = "he";
			} else if (AndroidPortAdditions.GAME_TYPE == GameType.IHNM)
			{
				// Languages appearing in IHNM
				if (checkLocale("it"))
					defaultLanguagePref = "it";
				if (checkLocale("es"))
					defaultLanguagePref = "es";
			}

			prefs.edit().putString("pref_language", defaultLanguagePref).commit();

			AndroidPortAdditions.instance().setFirstUse(false);

			MyLog.d("GameMenuActivity: onCreate: locale language: "
					+ Locale.getDefault().getLanguage());
			MyLog.d("GameMenuActivity: onCreate: default language pref: " + defaultLanguagePref);

			// Set the appropriate graphic mode or the device
			GraphicModeBehavior behavior = AndroidPortAdditions.instance().getGraphicModeBehavior();

			switch (behavior)
			{
			case SOFTWARE_SCALER:
				AndroidPortAdditions.instance().setScalingOption(
						AndroidPortAdditions.SCALING_OPTION_SOFT);
				AndroidPortAdditions.instance().setShaderTested(true);
				break;
			case ULTRA_HARDWARE_SCALER:
				AndroidPortAdditions.instance().setScalingOption(
						AndroidPortAdditions.SCALING_OPTION_ULTRA_SHADER);
				AndroidPortAdditions.instance().setShaderTested(true);
				break;
			case HARDWARE_SCALER:
				AndroidPortAdditions.instance().setScalingOption(
						AndroidPortAdditions.SCALING_OPTION_SHADER);
				AndroidPortAdditions.instance().setShaderTested(true);
				break;
			case TEST_HARDWARE_SCALER:
				AndroidPortAdditions.instance().setScalingOption(
						AndroidPortAdditions.SCALING_OPTION_SHADER);
				AndroidPortAdditions.instance().setShaderTested(false);
				break;
			case LQ_HARDWARE_SCALER:
				AndroidPortAdditions.instance().setScalingOption(
						AndroidPortAdditions.SCALING_OPTION_LQ_SHADER);
				AndroidPortAdditions.instance().setShaderTested(true);
				break;
			case TEST_LQ_HARDWARE_SCALER:
				AndroidPortAdditions.instance().setScalingOption(
						AndroidPortAdditions.SCALING_OPTION_LQ_SHADER);
				AndroidPortAdditions.instance().setShaderTested(false);
				break;
			case ORIGINAL:
				AndroidPortAdditions.instance().setScalingOption(
						AndroidPortAdditions.SCALING_OPTION_NONE);
				AndroidPortAdditions.instance().setShaderTested(true);
				break;
			}
		}

		// HACK: fix for a bug in Simon 1 which caused initialization with
		// non-existant languages
		if (AndroidPortAdditions.GAME_TYPE == GameType.SIMON1)
		{
			String langPref = prefs.getString("pref_language", "en");
			if (langPref.equals("cz") || langPref.equals("pl"))
			{
				prefs.edit().putString("pref_language", "en").commit();
			}
		}

		// Check if we need to set the default music mode (only one time)
		if (AndroidPortAdditions.instance().getFirstTimeSetMusicMode())
		{
			String currentMusicMode = prefs.getString("pref_music_mode", "original");

			MyLog.d("GameMenuActivity: onCreate: default music mode not set yet, current is "
					+ currentMusicMode);

			if (!currentMusicMode.equals("none"))
			{
				// If the current music mode is not "none" and the default was
				// never set, set it now
				String defaultMusicMode = "original";
				switch (AndroidPortAdditions.GAME_TYPE)
				{
				case SIMON1:
					defaultMusicMode = AndroidPortAdditions.instance().canUseMT32Emulation() ? "original_mt32"
							: "enhanced";
					break;
				case SIMON2:
				case FOTAQ:
					defaultMusicMode = AndroidPortAdditions.instance().canUseMT32Emulation() ? "original_mt32"
							: "original_eas";
					break;
				case T7G:
				case IHNM:
					defaultMusicMode = "enhanced";
				}

				prefs.edit().putString("pref_music_mode", defaultMusicMode).commit();

				MyLog.d("GameMenuActivity: onCreate: setting default music mode to "
						+ defaultMusicMode);

			}

			AndroidPortAdditions.instance().setFirstTimeSetMusicMode(false);
		}

		// Check if we need to set the default music volumes
		if (getFirstTimeSetVolumeSettings())
		{
			int speechVol = 192, sfxVol = 192;
			int musicVol = getDefaultMusicVolume();

			switch (AndroidPortAdditions.GAME_TYPE)
			{
			case SIMON1:
				speechVol = SIMON1_SPEECH_VOL;
				sfxVol = SIMON1_SFX_VOL;
				break;
			case SIMON2:
				speechVol = SIMON2_SPEECH_VOL;
				sfxVol = SIMON2_SFX_VOL;
				break;
			case FOTAQ:
				speechVol = FOTAQ_SPEECH_VOL;
				sfxVol = FOTAQ_SFX_VOL;
				break;
			}

			// Store the preferences
			prefs.edit().putInt(MUSIC_VOLUME_PREF, musicVol).putInt(SPEECH_VOLUME_PREF, speechVol)
					.putInt(SFX_VOLUME_PREF, sfxVol).commit();

			setFirstTimeSetVolumeSettings(false);
		}
	}
	
	public int getDefaultMusicVolume()
	{
		Context context = ScummVMApplication.getContext();

		SharedPreferences prefs = PreferenceManager.getDefaultSharedPreferences(context);
		String currentMusicMode = prefs.getString("pref_music_mode", "original");

		int musicVol = 192;
		if (currentMusicMode.equals("enhanced"))
		{
			// Only in Simon 1
		} else if (currentMusicMode.equals("original_mt32"))
		{
			switch (AndroidPortAdditions.GAME_TYPE)
			{
			case SIMON1:
				musicVol = SIMON1_MUSIC_VOL_MT32;
				break;
			case SIMON2:
				musicVol = SIMON2_MUSIC_VOL_MT32;
				break;
			case FOTAQ:
				musicVol = FOTAQ_MUSIC_VOL_MT32;
				break;
			}
		} else if (currentMusicMode.equals("original_eas"))
		{
			switch (AndroidPortAdditions.GAME_TYPE)
			{
			case SIMON1:
				musicVol = SIMON1_MUSIC_VOL_EAS;
				break;
			case SIMON2:
				musicVol = SIMON2_MUSIC_VOL_EAS;
				break;
			case FOTAQ:
				musicVol = FOTAQ_MUSIC_VOL_EAS;
				break;
			}
		} else if (currentMusicMode.equals("original"))
		{
			switch (AndroidPortAdditions.GAME_TYPE)
			{
			case SIMON1:
				musicVol = SIMON1_MUSIC_VOL_ADLIB;
				break;
			case SIMON2:
				musicVol = SIMON2_MUSIC_VOL_ADLIB;
				break;
			case FOTAQ:
				musicVol = FOTAQ_MUSIC_VOL_ADLIB;
				break;
			}
		}
		
		return musicVol;
	}

	public void germanVersionForceVoiceOnly(PreferenceManager prefManager)
	{

		SharedPreferences prefs = PreferenceManager.getDefaultSharedPreferences(ScummVMApplication
				.getContext());
		if (prefs.getString("pref_language", "en").equals("de")
				&& prefs.getString("pref_voice_mode", "voice_subtitles").equals("voice_subtitles"))
		{
			ListPreference pref = (ListPreference) prefManager.findPreference("pref_voice_mode");

			pref.setValue("voice");
		}
	}
	
	public void russianVersionForceVoiceOnly(PreferenceManager prefManager)
	{
		SharedPreferences prefs = PreferenceManager.getDefaultSharedPreferences(ScummVMApplication
				.getContext());
		if (prefs.getString("pref_language", "en").equals("ru"))
		{
			ListPreference pref = (ListPreference) prefManager.findPreference("pref_voice_mode");

			pref.setValue("voice");
		}
	}

	private boolean checkLocale(String languageCode)
	{
		// Compare the language code to the default locale's language code
		String locale = Locale.getDefault().getLanguage();
		MyLog.d("GameMenuActivity: checkLocale: locale is " + locale);
		return (locale.equals(new Locale(languageCode, "", "").getLanguage()));
	}

	public boolean getAutoGraphicMode()
	{
		return PreferenceManager.getDefaultSharedPreferences(ScummVMApplication.getContext())
				.getBoolean(AUTO_GRAPHIC_OPTION_PREF, true);
	}

	public void setAutoGraphicMode(boolean value)
	{
		PreferenceManager.getDefaultSharedPreferences(ScummVMApplication.getContext()).edit()
				.putBoolean(AUTO_GRAPHIC_OPTION_PREF, value).commit();
	}

	public int getScalingOption()
	{
		String scaleOption = PreferenceManager.getDefaultSharedPreferences(
				ScummVMApplication.getContext()).getString(SCALING_OPTION_PREF,
				"scaling_option_shader");

		if (scaleOption.equals("scaling_option_shader"))
		{
			return SCALING_OPTION_SHADER;
		}
		if (scaleOption.equals("scaling_option_shader_native_res"))
		{
			return SCALING_OPTION_ULTRA_SHADER;
		}
		if (scaleOption.equals("scaling_option_lq_shader"))
		{
			return SCALING_OPTION_LQ_SHADER;
		}
		if (scaleOption.equals("scaling_option_software"))
		{
			if (AndroidPortAdditions.SUPPORTS_SOFTWARE_SCALER)
			{
				return SCALING_OPTION_SOFT;
			} else
			{
				return SCALING_OPTION_NONE;
			}
		}
		if (scaleOption.equals("scaling_option_none"))
		{
			return SCALING_OPTION_NONE;
		}

		throw new IllegalStateException("invalid scaling option " + scaleOption);
	}

	public void setScalingOption(int option)
	{
		String scaleOption;
		switch (option)
		{
		case SCALING_OPTION_SHADER:
			scaleOption = "scaling_option_shader";
			break;
		case SCALING_OPTION_ULTRA_SHADER:
			scaleOption = "scaling_option_shader_native_res";
			break;
		case SCALING_OPTION_LQ_SHADER:
			scaleOption = "scaling_option_lq_shader";
			break;
		case SCALING_OPTION_SOFT:
			if (SUPPORTS_SOFTWARE_SCALER)
				scaleOption = "scaling_option_software";
			else
				scaleOption = "scaling_option_none";
			break;
		case SCALING_OPTION_NONE:
			scaleOption = "scaling_option_none";
			break;
		default:
			throw new IllegalStateException("invalid scaling option " + option);
		}

		PreferenceManager.getDefaultSharedPreferences(ScummVMApplication.getContext()).edit()
				.putString(SCALING_OPTION_PREF, scaleOption).commit();
	}

	public MusicOption getMusicOption()
	{
		String musicOption = PreferenceManager.getDefaultSharedPreferences(
				ScummVMApplication.getContext()).getString(MUSIC_OPTION_PREF,
				MusicOption.ADLIB.getValue());

		return MusicOption.fromValue(musicOption);
	}

	public int getSubtitleSpeed()
	{
		return PreferenceManager.getDefaultSharedPreferences(ScummVMApplication.getContext())
				.getInt(SUBTITLE_SPEED_PREF, 60);
	}

	public boolean getFirstUse()
	{
		return PreferenceManager.getDefaultSharedPreferences(ScummVMApplication.getContext())
				.getBoolean(FIRST_USE_PREF, true);
	}

	public void setFirstUse(boolean value)
	{
		Editor editor = PreferenceManager.getDefaultSharedPreferences(
				ScummVMApplication.getContext()).edit();
		editor.putBoolean(FIRST_USE_PREF, value);
		editor.commit();
	}

	public boolean getFirstTimeSetMusicMode()
	{
		return PreferenceManager.getDefaultSharedPreferences(ScummVMApplication.getContext())
				.getBoolean(FIRST_TIME_SET_MUSIC_MODE_PREF, true);
	}

	public void setFirstTimeSetMusicMode(boolean value)
	{
		Editor editor = PreferenceManager.getDefaultSharedPreferences(
				ScummVMApplication.getContext()).edit();
		editor.putBoolean(FIRST_TIME_SET_MUSIC_MODE_PREF, value);
		editor.commit();
	}

	public boolean getOpenHouseFlag()
	{
		if (T7G_ALWAYS_OPEN_HOUSE)
		{
			return true;
		}

		return PreferenceManager.getDefaultSharedPreferences(ScummVMApplication.getContext())
				.getBoolean(OPEN_HOUSE_PREF, false);
	}

	public void setOpenHouseFlag(boolean value)
	{
		Editor editor = PreferenceManager.getDefaultSharedPreferences(
				ScummVMApplication.getContext()).edit();
		editor.putBoolean(OPEN_HOUSE_PREF, value);
		editor.commit();
	}

	public boolean getEnablePopups()
	{
		return PreferenceManager.getDefaultSharedPreferences(ScummVMApplication.getContext())
				.getBoolean(ENABLE_POPUPS_PREF, true);
	}

	public void setEnablePopups(boolean value)
	{
		Editor editor = PreferenceManager.getDefaultSharedPreferences(
				ScummVMApplication.getContext()).edit();
		editor.putBoolean(ENABLE_POPUPS_PREF, value);
		editor.commit();
	}

	public boolean getFirstTimeSetVolumeSettings()
	{
		return PreferenceManager.getDefaultSharedPreferences(ScummVMApplication.getContext())
				.getBoolean(FIRST_TIME_SET_VOLUME_SETTINGS_PREF, true);
	}

	public void setFirstTimeSetVolumeSettings(boolean value)
	{
		Editor editor = PreferenceManager.getDefaultSharedPreferences(
				ScummVMApplication.getContext()).edit();
		editor.putBoolean(FIRST_TIME_SET_VOLUME_SETTINGS_PREF, value);
		editor.commit();
	}

	public boolean getShouldSwitchMusicFiles()
	{
		return PreferenceManager.getDefaultSharedPreferences(ScummVMApplication.getContext())
				.getBoolean(SHOULD_SWITCH_MUSIC_FILES_PREF, false);
	}

	public void setShouldSwitchMusicFiles(boolean value)
	{
		Editor editor = PreferenceManager.getDefaultSharedPreferences(
				ScummVMApplication.getContext()).edit();
		editor.putBoolean(SHOULD_SWITCH_MUSIC_FILES_PREF, value);
		editor.commit();
	}

	public boolean wasTutorialOffered()
	{
		return PreferenceManager.getDefaultSharedPreferences(ScummVMApplication.getContext())
				.getBoolean(OFFERED_TUTORIAL_PREF, false);
	}

	public void setTutorialOffered(boolean value)
	{
		Editor editor = PreferenceManager.getDefaultSharedPreferences(
				ScummVMApplication.getContext()).edit();
		editor.putBoolean(OFFERED_TUTORIAL_PREF, value);
		editor.commit();
	}

	public String getGlRendererPref()
	{
		return PreferenceManager.getDefaultSharedPreferences(ScummVMApplication.getContext())
				.getString(GL_RENDERER_PREF, "");
	}

	public void setGlRendererPref(String value)
	{
		Editor editor = PreferenceManager.getDefaultSharedPreferences(
				ScummVMApplication.getContext()).edit();
		editor.putString(GL_RENDERER_PREF, value);
		editor.commit();
	}

	public boolean getShaderTested()
	{
		if (TEST_SHADER_BENCHMARK)
		{
			// If this constant is set, always test the shader
			return false;
		}

		return PreferenceManager.getDefaultSharedPreferences(ScummVMApplication.getContext())
				.getBoolean(SHADER_TESTED_PREF, false);
	}

	public void setShaderTested(boolean value)
	{
		Editor editor = PreferenceManager.getDefaultSharedPreferences(
				ScummVMApplication.getContext()).edit();
		editor.putBoolean(SHADER_TESTED_PREF, value);
		editor.commit();
	}

	public String getAdGame()
	{
		return PreferenceManager.getDefaultSharedPreferences(ScummVMApplication.getContext())
				.getString(AD_POLICY_PREF, "");
	}

	public void setAdGame(String game)
	{
		Editor editor = PreferenceManager.getDefaultSharedPreferences(
				ScummVMApplication.getContext()).edit();
		editor.putString(AD_POLICY_PREF, game);
		editor.commit();
	}

	public boolean getGraphicSettingPromptShown()
	{
		return PreferenceManager.getDefaultSharedPreferences(ScummVMApplication.getContext())
				.getBoolean(GRAPHIC_SETTING_PROMPT_PREF, false);
	}

	public void setGraphicSettingPromptShown(boolean value)
	{
		Editor editor = PreferenceManager.getDefaultSharedPreferences(
				ScummVMApplication.getContext()).edit();
		editor.putBoolean(GRAPHIC_SETTING_PROMPT_PREF, value);
		editor.commit();
	}

	public int getLastSaveSlot()
	{
		return PreferenceManager.getDefaultSharedPreferences(ScummVMApplication.getContext())
				.getInt(LAST_SAVE_SLOT_PREF, -1);
	}

	public List<Integer> getAchievementIds()
	{
		return CsvUtils.csvToIntList(PreferenceManager.getDefaultSharedPreferences(
				ScummVMApplication.getContext()).getString(ACHIEVEMENTS_PREF, ""));
	}

	public boolean isAchievementUnlocked(int id)
	{
		return getAchievementIds().contains(id);
	}

	public void setAchievementIds(List<Integer> ids)
	{
		MyLog.d("AndroidPortAdditions: setAchievementIds: " + ids);

		Editor editor = PreferenceManager.getDefaultSharedPreferences(
				ScummVMApplication.getContext()).edit();
		editor.putString(ACHIEVEMENTS_PREF, CsvUtils.intListToCsv(ids));
		editor.commit();
	}

	public void addAchievementIds(List<Integer> ids)
	{
		MyLog.d("AndroidPortAdditions: addAchievementIds: " + ids);

		// Use a set to contain only one value for each achievement
		List<Integer> currentIds = getAchievementIds();
		TreeSet<Integer> idSet = new TreeSet<Integer>(currentIds);
		idSet.addAll(ids);

		List<Integer> finalIds = new ArrayList<Integer>(idSet);

		MyLog.d("AndroidPortAdditions: addAchievementIds: final ids: " + finalIds);

		Editor editor = PreferenceManager.getDefaultSharedPreferences(
				ScummVMApplication.getContext()).edit();
		editor.putString(ACHIEVEMENTS_PREF, CsvUtils.intListToCsv(finalIds));
		editor.commit();
	}

	public boolean addAchievementId(int id)
	{
		// Add the id if needed
		List<Integer> achievementIds = getAchievementIds();
		if (!achievementIds.contains(id))
		{
			achievementIds.add(id);

			// Commit to preferences
			Editor editor = PreferenceManager.getDefaultSharedPreferences(
					ScummVMApplication.getContext()).edit();
			editor.putString(ACHIEVEMENTS_PREF, CsvUtils.intListToCsv(achievementIds));
			editor.commit();

			MyLog.d("AndroidPortAdditions: addAchievementId: unlocked new achievement: " + id);

			return true;
		}

		return false;
	}

	public void clearAllAchievements()
	{
		Editor editor = PreferenceManager.getDefaultSharedPreferences(
				ScummVMApplication.getContext()).edit();
		editor.putString(ACHIEVEMENTS_PREF, "");
		editor.commit();
	}

	/**
	 * 
	 * @return List of save slot names
	 */
	public List<String> getSaveSlots()
	{
		// Get the save slots preference as a list of strings
		return CsvUtils.csvToStringList(PreferenceManager.getDefaultSharedPreferences(
				ScummVMApplication.getContext()).getString(SAVE_SLOTS_PREF, sDefaultSaveSlots));
	}

	/**
	 * 
	 * @return 0-based indexes of string list positions
	 */
	public List<Integer> getSaveSlotsIndexes()
	{
		// Get the save slots preference as a list of strings
		return CsvUtils.csvToIntList(PreferenceManager.getDefaultSharedPreferences(
				ScummVMApplication.getContext()).getString(SAVE_SLOTS_INDEXES_PREF, ""));
	}

	/**
	 * 
	 * @param slot
	 *            1-based save slot according to the game engine
	 * @param name
	 */
	public void setSaveSlot(int slot, String name)
	{
		if (slot < 1 || slot > MAX_SAVE_SLOTS)
		{
			MyLog.e("Illegal save slot, must be 0 - " + MAX_SAVE_SLOTS);
			return;
		}

		// Modify the slot to the chosen name
		List<String> saveSlots = getSaveSlots();
		saveSlots.set(slot - 1, name);

		// Add the slot index if needed
		Integer index = Integer.valueOf(slot - 1);
		List<Integer> slotIndexes = getSaveSlotsIndexes();
		if (!slotIndexes.contains(index))
		{
			slotIndexes.add(index);
		}

		// Commit to preferences
		Editor editor = PreferenceManager.getDefaultSharedPreferences(
				ScummVMApplication.getContext()).edit();
		editor.putString(SAVE_SLOTS_PREF, CsvUtils.stringListToCsv(saveSlots));
		editor.putString(SAVE_SLOTS_INDEXES_PREF, CsvUtils.intListToCsv(slotIndexes));
		// Update the last save slot
		editor.putInt(LAST_SAVE_SLOT_PREF, slot);
		editor.commit();
	}

	public long getGameFilesSize()
	{
		return PreferenceManager.getDefaultSharedPreferences(ScummVMApplication.getContext())
				.getLong(GAME_FILES_SIZE_PREF, -1);
	}

	public void setGameFilesSize(long value)
	{
		Editor editor = PreferenceManager.getDefaultSharedPreferences(
				ScummVMApplication.getContext()).edit();
		editor.putLong(GAME_FILES_SIZE_PREF, value);
		editor.commit();
	}

	public String getGameFileDir()
	{
		return sGameFileDir;
	}

	public void setGameFileParentDir(String parentDir)
	{
		sGameFileDir = parentDir + GAME_FILE_DIR;
	}

	public int getAPKExtensionFileVersion()
	{
		return PreferenceManager.getDefaultSharedPreferences(ScummVMApplication.getContext())
				.getInt(LATEST_OBB_PREF, -1);
	}

	public void setAPKExtensionFileVersion(int version)
	{
		// Check if we need to reset the game files
		int oldVersion = getAPKExtensionFileVersion();
		if (version > oldVersion)
		{
			AndroidPortAdditions.instance().setGameFilesSize(-1);
		}
		
		Editor editor = PreferenceManager.getDefaultSharedPreferences(
				ScummVMApplication.getContext()).edit();
		editor.putInt(LATEST_OBB_PREF, version);
		editor.commit();
	}

	/**
	 * Used for keeping track of whether the game activity is alive in this
	 * process
	 * 
	 * @return
	 */
	public boolean isGameActivityAlive()
	{
		return (mGameActivityCount > 0);
	}

	public void onGameActivityCreate()
	{
		++mGameActivityCount;
	}

	public void onGameActivityDestroy()
	{
		--mGameActivityCount;
	}

	public boolean checkUltraModeSupport()
	{
		String model = Build.MODEL;
		// Check if the device is in our ultra mode supported models
		for (String s : DeviceConfigurations.instance().getDevicesUsingUltraMode())
		{
			if (s.equalsIgnoreCase(model))
			{
				return true;
			}
		}

		String deviceGlRenderer = getGlRendererPref();
		// Check if the device has a GPU that supports ultra mode all the time
		for (String s : DeviceConfigurations.instance().getGPUKindsUsingUltraMode())
		{
			if (s.equalsIgnoreCase(deviceGlRenderer))
			{
				return true;
			}
		}

		return false;
	}

	public boolean canUseMT32Emulation()
	{
		DeviceModel model = DeviceIdentifier.getDeviceModel();

		// Check if the device is in our Mt32 supported models
		for (DeviceModel m : DeviceConfigurations.instance().getDevicesUsingMT32Emu())
		{
			if (m.equals(model))
			{
				return true;
			}
		}

		return false;
	}

	public GraphicModeBehavior getGraphicModeBehavior()
	{
		String model = Build.MODEL;
		String deviceGlRenderer = getGlRendererPref();

		MyLog.d("AndroidPortAdditions: getGraphicModeBehavior: model " + model);
		MyLog.d("AndroidPortAdditions: getGraphicModeBehavior: GL_RENDERER " + deviceGlRenderer);

		// Send statistics of device and GPU
		StatisticsManager.getStatistics().reportDeviceAndGPU(model, deviceGlRenderer);

		// If the device supports ultra mode, it automatically uses the 'high'
		// mode
		if (checkUltraModeSupport())
		{
			return GraphicModeBehavior.ULTRA_HARDWARE_SCALER;
		}

		// Check if the device is in our hardware scaler supported models
		for (String s : DeviceConfigurations.instance().getDevicesUsingHardwareScaler())
		{
			if (s.equalsIgnoreCase(model))
			{
				// Check if the device also has a supported GPU
				for (String s2 : DeviceConfigurations.instance()
						.getGPUKindsUsingHardwareScalerTest())
				{
					if (s2.equalsIgnoreCase(deviceGlRenderer))
					{
						return GraphicModeBehavior.HARDWARE_SCALER;
					}
				}
			}
		}

		// Check if the device is in our LQ hardware scaler supported models
		for (String s : DeviceConfigurations.instance().getDevicesUsingLQHardwareScaler())
		{
			if (s.equalsIgnoreCase(model))
			{
				return GraphicModeBehavior.LQ_HARDWARE_SCALER;
			}
		}

		// Check if the device GL_RENDERER (GPU) is in our HW test list
		for (String s : DeviceConfigurations.instance().getGPUKindsUsingHardwareScalerTest())
		{
			if (s.equalsIgnoreCase(deviceGlRenderer))
			{
				return GraphicModeBehavior.TEST_HARDWARE_SCALER;
			}
		}

		// Check if the device GL_RENDERER (GPU) is in our LQ list (no test)
		for (String s : DeviceConfigurations.instance().getGPUKindsUsingLQHardwareScaler())
		{
			if (s.equalsIgnoreCase(deviceGlRenderer))
			{
				return GraphicModeBehavior.LQ_HARDWARE_SCALER;
			}
		}

		// Check if the GPU is in the SW scaler list
		for (String s : DeviceConfigurations.instance().getGPUKindsUsingSoftwareScaler())
		{
			if (s.equalsIgnoreCase(deviceGlRenderer))
			{
				if (GAME_TYPE == GameType.T7G || GAME_TYPE == GameType.IHNM)
					return GraphicModeBehavior.ORIGINAL;
				else
					return GraphicModeBehavior.SOFTWARE_SCALER;
			}
		}

		// Default to LQ hardware scaler
		return GraphicModeBehavior.TEST_LQ_HARDWARE_SCALER;
	}
	
	public void setStartNewGameAltIntro(boolean value)
	{
		mStartNewGameAltIntro = value;
	}
	
	public boolean getStartNewGameAltIntro()
	{
		return mStartNewGameAltIntro;
	}

	private AndroidPortAdditions()
	{

	}

	private static AndroidPortAdditions sInstance;

	private String sGameFileDir;

	private int sExtensionFileVersion;

	private int mGameActivityCount = 0;
	
	private boolean mStartNewGameAltIntro = false;

	private static final String OFFERED_TUTORIAL_PREF = "offered_tutorial_pref";

	private static final String FIRST_USE_PREF = "first_use_pref";
	private static final String SHADER_TESTED_PREF = "shader_tested_pref";
	private static final String GRAPHIC_SETTING_PROMPT_PREF = "grapihc_setting_prompt_pref";

	private static final String GAME_FILES_SIZE_PREF = "game_files_size_pref";
	private static final String ACHIEVEMENTS_PREF = "acheievement_pref";

	private static final String SAVE_SLOTS_PREF = "save_slots_pref";
	private static final String SAVE_SLOTS_INDEXES_PREF = "save_slots_indexes_pref";
	private static final String LAST_SAVE_SLOT_PREF = "last_save_slot_pref";

	private static final String AUTO_GRAPHIC_OPTION_PREF = "pref_auto_graphic_option";

	private static final String SCALING_OPTION_PREF = "pref_scaler_option";

	private static final String GL_RENDERER_PREF = "pref_gl_renderer";

	private static final String AD_POLICY_PREF = "pref_ad_policy";

	private static final String FIRST_TIME_SET_MUSIC_MODE_PREF = "first_time_set_music_mode_pref";
	private static final String FIRST_TIME_SET_VOLUME_SETTINGS_PREF = "first_time_set_volume_settings_pref";

	private static final String SHOULD_SWITCH_MUSIC_FILES_PREF = "should_switch_music_files_pref";

	private static final String OPEN_HOUSE_PREF = "open_house_pref";
	private static final String ENABLE_POPUPS_PREF = "enable_popups_pref";

	private static final String MUSIC_OPTION_PREF = "pref_music_mode";

	private static final String MUSIC_VOLUME_PREF = "pref_music_volume";
	private static final String SPEECH_VOLUME_PREF = "pref_speech_volume";
	private static final String SFX_VOLUME_PREF = "pref_sfx_volume";

	private static final String SUBTITLE_SPEED_PREF = "pref_subtitle_speed";

	private static final String LATEST_OBB_PREF = "latest_obb_pref";

	private static final int SIMON1_MUSIC_VOL_NEW_RECORDING = 192;
	private static final int SIMON1_MUSIC_VOL_MT32 = 255;
	private static final int SIMON1_MUSIC_VOL_EAS = 150;
	private static final int SIMON1_MUSIC_VOL_ADLIB = 140;

	private static final int SIMON2_MUSIC_VOL_MT32 = 255;
	private static final int SIMON2_MUSIC_VOL_EAS = 150;
	private static final int SIMON2_MUSIC_VOL_ADLIB = 140;

	private static final int FOTAQ_MUSIC_VOL_MT32 = 255;
	private static final int FOTAQ_MUSIC_VOL_EAS = 150;
	private static final int FOTAQ_MUSIC_VOL_ADLIB = 140;

	private static final int SIMON1_SPEECH_VOL = 192;
	private static final int SIMON2_SPEECH_VOL = 192;
	private static final int FOTAQ_SPEECH_VOL = 192;

	private static final int SIMON1_SFX_VOL = 192;
	private static final int SIMON2_SFX_VOL = 100;
	private static final int FOTAQ_SFX_VOL = 100;

	private static final int MAX_SAVE_SLOTS = 8;

	// Initialize the default formation of save slots
	private static String sDefaultSaveSlots;
	static
	{
		ArrayList<String> saveSlots = new ArrayList<String>();
		for (int i = 0; i < MAX_SAVE_SLOTS; ++i)
		{
			// Adjust the default slot ordinals, based on whether we support
			// auto-save or not
			int slotNameOrdinal = SUPPORTS_AUTO_SAVE ? i : i + 1;

			saveSlots.add("EMPTY SLOT " + String.valueOf(slotNameOrdinal));
		}

		sDefaultSaveSlots = CsvUtils.stringListToCsv(saveSlots);
	}

}
